(*
  Connects Lexer with Parser and prints reasonable error messages.

  author: Christian Mueller
*)

exception LexerParserError 

(* exception handling for lexer and parser *)

fun errorHandler (Parsing.ParseError f) instream lexbuf =
    let
        val pos1 = Lexing.getLexemeStart lexbuf
        val pos2 = Lexing.getLexemeEnd lexbuf
    in
        Location.errMsg ("",instream,lexbuf)
                        (Location.Loc(pos1, pos2))
                        "Syntax Error!"
			handle Fail s => raise LexerParserError
    end
  | errorHandler (Lexer.LexerError(msg,pos1,pos2)) instream lexbuf =
    (if pos1 >= 0 andalso pos2 >= 0 then
         Location.errMsg ("", instream, lexbuf)
                         (Location.Loc(pos1, pos2))
                         ("Lexical error: " ^ msg)
			 handle Fail s => raise LexerParserError
     else
         (Location.errPrompt ("Lexical error: " ^ msg ^ "\n\n");
          raise LexerParserError))
  | errorHandler exn _ _ = raise exn

(* parse file *)

fun createLexerStream is = 
    Lexing.createLexer (fn buff => fn n => Nonstdio.buff_input is buff 0 n)

fun parse instream parser lexer lexbuf=
    let
        val expr = parser lexer lexbuf 
	    handle e => errorHandler e instream lexbuf
    in
        Parsing.clearParser();
        expr nil
    end
    handle Support.LabelError(pos1,pos2) =>   
	          (Parsing.clearParser();
		   Location.errMsg ("", instream, lexbuf)
		   (Location.Loc(pos1, pos2))
		   "The same label is bound twice"
		   handle Fail s => raise LexerParserError)
	 | exn => (Parsing.clearParser(); raise exn);

(* lexer and parser for Lambda Cube Toplevel *)

fun Parser stream = 
    parse stream Parser.Toplevel Lexer.Token (createLexerStream stream)

fun Fileparser stream = 
    parse stream Parser.ToplevelFile Lexer.Token (createLexerStream stream)

(* parses with the given parser and returns one huge command list *)

fun parse_files files =
    let
	fun load file = (Nonstdio.open_in_bin file 
			 handle _ => raise Fail("file \""^file^"\" not found"))

	fun process  nil       acc = acc
	  | process (file::xr) acc = 
	    let
		val stream = load file
		(* stores info for error reporting *)
		val _ = Info.setCurrentFile file 

		val acc' = Fileparser stream @ acc
	    in
		BasicIO.close_in stream;
		process xr acc'
	    end	
    in
	process files nil
    end


