(* 
   This files implements graph based type reconstruction. 
   As the normal representation is in tree-form, there is an algorithm 
   that transforms the tree of the abstract syntax into a corresponding graph.
   There is also the transformation from graphs to trees available, so that 
   you can use the PrettyPrinter.

   Normally, these algorithms avoid the exponential complexity of the syntax 
   based methods. However macro expension and cycle checks might destroy 
   this property.
   
   author: Christian Mueller    
*)

local
    open AbstractSyntax LambdaCube LCSupport PrettyPrinter Info
in

(* 
 Types are represented as (cyclic) graphs.
 They are represented as a dynamic array that can be modified with the 
 following basic operations
*)

type typ = int
datatype info = S | V | INT | BOOL | UNIT | EMPTY |  A of typ * typ | F of typ

val graph = Dynarray.array(50, EMPTY) 
val nextnode = ref 0

fun get t = Dynarray.sub(graph,t)
fun put t i = Dynarray.update(graph,t,i)
fun new i = let val t = !nextnode in nextnode:=t+1; put t i; t end
fun cleargraph () = 
    let
	val size = Dynarray.bound graph
	fun loop n = if n < size then (put n EMPTY; loop(n+1)) else ()
    in
	loop 0
    end

fun pr () = (* only needed for debugging *)
    let 
	val c = ref 0
	fun inc c = !c before c:=(!c)+1

	fun pr' S = print ("["^Int.toString (inc c)^".*] ")
	  | pr' V = print ("["^Int.toString (inc c)^".V] ")
	  | pr' INT = print ("["^Int.toString (inc c)^".INT] ")
	  | pr' BOOL = print ("["^Int.toString (inc c)^".BOOL] ")
	  | pr' UNIT = print ("["^Int.toString (inc c)^".UNIT] ")
	  | pr' EMPTY = (*print ("["^Int.toString (inc c)^".EMPTY] ")*) ()
	  | pr' (F n) = 
	    print ("["^Int.toString (inc c)^".F "^Int.toString n^"] ")

	  | pr' (A(n1,n2)) = 
	    print ("["^Int.toString (inc c)^".A("^Int.toString n1^","
		   ^Int.toString n2^")] ")

	val size = Dynarray.bound graph 

	fun loop n = if n < size then (pr' (get n); loop (n+1)) else ()
    in
	print "graph = "; loop 0; print "\n"
    end

exception Cycle
exception UnifyError of LCSupport.mixture list
exception ReconGBError of string

(* cycle checks: tests whether the first arg ist already in the list xs *)

fun cycle t xs = 
    case List.find (fn t' => t=t') xs of 
	NONE => false 
      | SOME _ => true

fun examined (t1,t2) xs = 
    case List.find (fn (t1',t2') => t1=t1' andalso t2=t2') xs of
	NONE => false
      | SOME _ => true

(*
 These functions transform a graph into a tree so that the type can be used by
 all other methods of the interpreter. There is some bookkeeping of free names
 to allow canonical output.
*)

(* functions for natural choice of variable names *)

val polynames = ref nil : (int*string) list ref
val polycount = ref 0

val tvnames = ref nil : (int*string) list ref
val tvcount = ref 0

fun findname x nil = raise Subscript
  | findname x ((y,name)::yr) = if x=y then name else findname x yr
    
fun getpolyvar x = 
    let 
	fun f x s = 
	    if x <= 0 then (if s="" then "a" else s)
	    else f (x div 26) (s^Char.toString(chr((x mod 26)+ord(#"a"))))

	val name = "'"^ f (!polycount) ""
    in
	findname x (!polynames) 
	handle Subcript => (polycount:=(!polycount)+1;
			    polynames := (x,name) :: (!polynames); name)
    end

fun gettypevar x = 
    let 
	fun f x s = 
	    if x <= 0 then (if s="" then "A" else s)
	    else f (x div 26) (s^Char.toString(chr((x mod 26)+ord(#"A"))))

	val name = f (!tvcount) ""
    in
	findname x (!tvnames) 
	handle Subcript => (tvcount:=(!tvcount)+1;
			    tvnames := (x,name) :: (!tvnames); name)
    end

(* 
 function to transform the (connected) graph starting at "t" into a tree; 
 a cycle check is performed to handle recursive types 
*)

fun toT t ex = 
    if cycle t ex then raise Cycle
    else (case get t of 
	      V => Var (emptyInfo,getpolyvar t,0) 
	    | INT => IntType emptyInfo
	    | BOOL => BoolType emptyInfo
	    | UNIT => UnitType emptyInfo
	    | EMPTY => raise Fail "wrong built graph"
	    | S => Star emptyInfo
	    | A(t1,t2) =>  
		  let			  
		      val (T1,v1) = (toT t1 (t::ex),"") 
			  handle Cycle => (Var (emptyInfo,gettypevar t1,t1), 
					   gettypevar t1) 

		      val (T2,v2) = (toT t2 (t::ex),"") 
			  handle Cycle => (Var (emptyInfo,gettypevar t2,t2), 
					   gettypevar t2) 

		      val T = Pi(emptyInfo,"",T1,T2) 
		  in			     
		      case (v1,v2) of 
			  ("","") => T 
			| ("",s2) => RecursiveSort(emptyInfo,s2,T) 
			| (s1,"") => RecursiveSort(emptyInfo,s1,T) 
			| (s1,s2) => 
			  if s1=s2 then RecursiveSort(emptyInfo,s1,T) 
			  else RecursiveSort(emptyInfo,s1,
					     RecursiveSort(emptyInfo,s2,T))
		  end
	    | F t' => toT t' (t::ex) )

fun toTree t = toT t nil

(*
 function to transform a tree into a graph;
 these is needed to handle type anotations properly
*)		       

fun fromTree tr =
    let
	fun raise_list xs = List.map (fn (e,n) => (shift 0 1 e,n)) xs

	fun getnode T xs = (* throws Option *)
	    #2 (valOf (List.find (fn (T',_) => equiv (T,T')) xs)) 
		   
	fun fromT (IntType i)  lst = basetype (IntType i)  INT lst
	  | fromT (BoolType i) lst = basetype (BoolType i) BOOL lst
	  | fromT (UnitType i) lst = basetype (UnitType i) UNIT lst
	  | fromT (Star i)     lst = 
	    let 
		val nn = new S 
	    in 
		(nn,((Star i,nn)::lst)) 
	    end

	  | fromT (v as Var (_,_,x)) lst = basetype v V lst
          | fromT (p as Pi(_,_,t1,t2)) lst =
	    (let val nn = getnode p lst in (nn,(p,nn)::lst) end
		 handle Option =>  
			let 
			    val nn = new V 
			    val (n1,lst') = fromT t1 ((p,nn)::lst) 
			    val (n2,lst'')= fromT t2 (raise_list lst') 
			in
			    put nn (A(n1,n2)); 
			    (nn,lst'') 
			end)  
	  | fromT (r as RecursiveSort(_,s,t)) lst = 
	    let 
		val nn = new V
		val (n,lst') = 
		    fromT t ((Var (emptyInfo,"V",0),nn)
			     ::(raise_list ((r,nn)::lst)))
	    in
		put nn (F n);
		(nn,lst')
	    end
          | fromT _ _ = raise Fail "fromTree: not supported" 

	and basetype T t lst = 
	    let 
		val nn = getnode T lst handle Option => new t
	    in 
		(nn,(T,nn)::lst)
	    end
    in
	#1 (fromT tr nil)
    end

(*
 Type reconstruction:
 Here is the type reconstruction algorithm. All functions can handle cycles 
 which is important for equi-recursive types.
*)

(* Follows forward nodes. Raises an exception if a cycle is found *)
fun deref t = 
    let 
	fun d t ex = 
	    if cycle t ex then raise Cycle 
	    else case get t of F t' => d t' (t::ex) | _ => t
    in 
	d t nil
    end

(* unification *)

fun unify nil _ = ()
  | unify ((t1,t2)::xr) ex = 
    let
	val (t1',t2') = (deref t1, deref t2) 	
	val ex' = (t1',t2')::ex
    in
	if t1'=t2' orelse examined (t1',t2') ex then unify xr ex'
	else case (get t1',get t2') of
	         (V,i) => (put t1' (F t2'); unify xr ex')
	       | (i,V) => (put t2' (F t1'); unify xr ex')
	       | (A(t11,t12),A(t21,t22)) => 
		     unify ((t11,t21)::(t12,t22)::xr) ex'

	       | (INT,INT) => unify xr ex'
	       | (BOOL,BOOL) => unify xr ex'
	       | (UNIT,UNIT) => unify xr ex'
	       | (S,S) => unify xr ex'
	       | _ => raise UnifyError ([Str "could not unify ", 
					 E (toTree t1'), Str " and ", 
					 E (toTree t2')])
    end

(* 
 constraint typing relation
 The constraints are written into the graph. The algorithm is interwined with 
 unifiction so that constraints are directly solved when they are set.
*)

fun recon s (Var (_,_,x)) = (List.nth(s,x) handle Subscript => new S)
  | recon s (Unit _)    = new UNIT
  | recon s (True _)    = new BOOL
  | recon s (False _)   = new BOOL
  | recon s (Int _) = new INT
  | recon s (If(_,e1,e2,e3)) = 
    let 
	val (t1,t2,t3) = (recon s e1, recon s e2, recon s e3)
    in  
	unify [(t2,t3),(t1,new BOOL)] nil; t2
    end
  | recon s (Plus(_,e1,e2))        = reconbinop s e1 e2 INT INT
  | recon s (Minus(_,e1,e2))       = reconbinop s e1 e2 INT INT
  | recon s (Mult(_,e1,e2))        = reconbinop s e1 e2 INT INT
  | recon s (IntEqual(_,e1,e2))    = reconbinop s e1 e2 INT BOOL
  | recon s (IntNotEqual(_,e1,e2)) = reconbinop s e1 e2 INT BOOL
  | recon s (LessEq(_,e1,e2))      = reconbinop s e1 e2 INT BOOL
  | recon s (GreaterEq(_,e1,e2))   = reconbinop s e1 e2 INT BOOL
  | recon s (BoolNeg (_,e)) = 
    let 
	val (n,ne) = (new BOOL, recon s e) 
    in 
	unify [(n,ne)] nil; n 
    end

  | recon s (IntNeg (_,e))  = 
    let 
	val (n,ne) = (new INT, recon s e) 
    in 
	unify [(n,ne)] nil; n 
    end

  | recon s (Abstr(_,_,Box _,e))   = 
    let 
	val v = new V 
    in 
	new (A(v, (recon (v::s) e))) 
    end

  | recon s (Abstr(_,_,T,e))     = 
    let 
	val _ = unify [(recon s T,new S)] nil
	val t = fromTree T 					       
    in  
	new (A(t, (recon (t::s) e))) 
    end

  | recon s (App(_,e1,e2)) = 
    let
	val (t1,t2) = (recon s e1, recon s e2)
	val x = new V
    in
	unify [(t1,new (A(t2,x)))] nil ; x
    end

  (* types need not to be checked since explicit syntax does not allow 
     higher-ordered types *)
  | recon s (Pi (_,_,e1,e2)) = 
    let 
	val k1 = recon s e1
	val k2 = recon (k1::s) e2
	val star = new S
    in 
	unify [(k1,star),(k2,star)] nil; star
    end

  | recon _ (UnitType i) = new S 
  | recon _ (BoolType i) = new S 
  | recon _ (IntType i) = new S 
  | recon s (RecursiveSort (_,_,t)) = (unify [(recon s t,new S)] nil; new S )
  | recon _ _ = raise ReconGBError "construct not supported" 

and reconbinop s e1 e2 argtype resulttype =
    let 
	val (t1,t2) = (recon s e1, recon s e2)		      
	val argtype_node = new argtype
	val resulttype_node = if argtype = resulttype then argtype_node 
			      else new resulttype
    in
	unify [(t1,argtype_node),(t2,argtype_node)] nil; resulttype_node
    end

(* resets all imperative parameters *)

fun clear () = (cleargraph ();
		nextnode := 0;
		polynames := nil;
		tvnames := nil;
		polycount := 0;
		tvcount := 0)
		    
(* 
 the type checking function;
 first resets parameters, then checks the type of e 
 and finally returns a tree that represents the type of e
*)
	  
fun check e = (	clear (); toTree (recon nil e) )

(* 

(* check with debugging output  *) 
fun check e = (* checks and outputs the array *)
    let
	val t = (clear (); recon nil e)
(*	val t = (clear(); 
		 fromTree (RecursiveSort(emptyInfo,"X",
					 RecursiveSort(emptyInfo,"Y",
						       Pi(emptyInfo,"",
							  Var(emptyInfo,"X",1),
							  Var(emptyInfo,"Y",1))))))
 *)
    in
	pr (); 
	print ("type starts at node "^Int.toString t^"\n");
	toTree t 
    end
*)

end 
